from .mergetree import (
    MergeTree, AggregatingMergeTree, CollapsingMergeTree,
    VersionedCollapsingMergeTree, ReplacingMergeTree, SummingMergeTree
)
from .util import parse_columns


class ReplicatedEngineMixin(object):
    def __init__(self, table_path, replica_name):
        self.table_path = table_path
        self.replica_name = replica_name

    def get_parameters(self):
        return [
            "'{}'".format(self.table_path),
            "'{}'".format(self.replica_name)
        ]

    @classmethod
    def _reflect_replicated(cls, engine_full):
        engine = parse_columns(engine_full, delimeter=' ')[0]
        columns = engine[len(cls.__name__):][1:-1]
        return parse_columns(columns)


class ReplicatedMergeTree(ReplicatedEngineMixin, MergeTree):
    def __init__(self, table_path, replica_name,
                 *args, **kwargs):
        ReplicatedEngineMixin.__init__(self, table_path, replica_name)
        MergeTree.__init__(self, *args, **kwargs)

    def get_parameters(self):
        return self.extend_parameters(
            ReplicatedEngineMixin.get_parameters(self),
            MergeTree.get_parameters(self)
        )

    @classmethod
    def reflect(cls, engine_full, **kwargs):
        args = cls._reflect_replicated(engine_full)
        table_path, replica_name = args[:2]

        return cls(
            table_path.strip("'"), replica_name.strip("'"),
            *args[2:],
            **cls._reflect_merge_tree(**kwargs)
        )


class ReplicatedAggregatingMergeTree(ReplicatedEngineMixin,
                                     AggregatingMergeTree):
    def __init__(self, table_path, replica_name,
                 *args, **kwargs):
        ReplicatedEngineMixin.__init__(self, table_path, replica_name)
        AggregatingMergeTree.__init__(self, *args, **kwargs)

    def get_parameters(self):
        return self.extend_parameters(
            ReplicatedEngineMixin.get_parameters(self),
            AggregatingMergeTree.get_parameters(self)
        )

    @classmethod
    def reflect(cls, engine_full, **kwargs):
        args = cls._reflect_replicated(engine_full)
        table_path, replica_name = args[:2]

        return cls(
            table_path.strip("'"), replica_name.strip("'"),
            *args[2:],
            **cls._reflect_merge_tree(**kwargs)
        )


class ReplicatedCollapsingMergeTree(ReplicatedEngineMixin,
                                    CollapsingMergeTree):
    def __init__(self, table_path, replica_name,
                 *args, **kwargs):
        ReplicatedEngineMixin.__init__(self, table_path, replica_name)
        CollapsingMergeTree.__init__(self, *args, **kwargs)

    def get_parameters(self):
        return self.extend_parameters(
            ReplicatedEngineMixin.get_parameters(self),
            CollapsingMergeTree.get_parameters(self)
        )

    @classmethod
    def reflect(cls, engine_full, **kwargs):
        args = cls._reflect_replicated(engine_full)
        table_path, replica_name = args[:2]

        return cls(
            table_path.strip("'"), replica_name.strip("'"),
            *args[2:],
            **cls._reflect_merge_tree(**kwargs)
        )


class ReplicatedVersionedCollapsingMergeTree(ReplicatedEngineMixin,
                                             VersionedCollapsingMergeTree):
    def __init__(self, table_path, replica_name,
                 *args, **kwargs):
        ReplicatedEngineMixin.__init__(self, table_path, replica_name)
        VersionedCollapsingMergeTree.__init__(self, *args, **kwargs)

    def get_parameters(self):
        return self.extend_parameters(
            ReplicatedEngineMixin.get_parameters(self),
            VersionedCollapsingMergeTree.get_parameters(self)
        )

    @classmethod
    def reflect(cls, engine_full, **kwargs):
        args = cls._reflect_replicated(engine_full)
        table_path, replica_name = args[:2]

        return cls(
            table_path.strip("'"), replica_name.strip("'"),
            *args[2:],
            **cls._reflect_merge_tree(**kwargs)
        )


class ReplicatedReplacingMergeTree(ReplicatedEngineMixin, ReplacingMergeTree):
    def __init__(self, table_path, replica_name,
                 *args, **kwargs):
        ReplicatedEngineMixin.__init__(self, table_path, replica_name)
        ReplacingMergeTree.__init__(self, *args, **kwargs)

    def get_parameters(self):
        return self.extend_parameters(
            ReplicatedEngineMixin.get_parameters(self),
            ReplacingMergeTree.get_parameters(self)
        )

    @classmethod
    def reflect(cls, engine_full, **kwargs):
        args = cls._reflect_replicated(engine_full)
        table_path, replica_name = args[:2]

        return cls(
            table_path.strip("'"), replica_name.strip("'"),
            *args[2:],
            **cls._reflect_merge_tree(**kwargs)
        )


class ReplicatedSummingMergeTree(ReplicatedEngineMixin, SummingMergeTree):
    def __init__(self, table_path, replica_name,
                 *args, **kwargs):
        ReplicatedEngineMixin.__init__(self, table_path, replica_name)
        SummingMergeTree.__init__(self, *args, **kwargs)

    def get_parameters(self):
        return self.extend_parameters(
            ReplicatedEngineMixin.get_parameters(self),
            SummingMergeTree.get_parameters(self)
        )

    @classmethod
    def reflect(cls, engine_full, **kwargs):
        args = cls._reflect_replicated(engine_full)
        table_path, replica_name = args[:2]
        columns = None
        if len(args) > 2:
            columns = tuple(parse_columns(args[2].strip('()')))

        return cls(
            table_path.strip("'"), replica_name.strip("'"),
            columns=columns,
            **cls._reflect_merge_tree(**kwargs)
        )



from sqlalchemy.util import to_list

from .base import Engine, KeysExpressionOrColumn, TableCol
from .util import parse_columns


class MergeTree(Engine):

    __visit_name__ = 'merge_tree'

    def __init__(
            self,
            partition_by=None,
            order_by=None,
            primary_key=None,
            sample_by=None,
            ttl=None,
            **settings
    ):
        self.partition_by = None
        if partition_by is not None:
            self.partition_by = KeysExpressionOrColumn(*to_list(partition_by))

        self.order_by = None
        if order_by is not None:
            self.order_by = KeysExpressionOrColumn(*to_list(order_by))

        self.primary_key = None
        if primary_key is not None:
            self.primary_key = KeysExpressionOrColumn(*to_list(primary_key))

        self.sample_by = None
        if sample_by is not None:
            self.sample_by = KeysExpressionOrColumn(sample_by)

        self.ttl = None
        if ttl is not None:
            self.ttl = KeysExpressionOrColumn(*to_list(ttl))

        self.settings = settings
        super(MergeTree, self).__init__()

    def _set_parent(self, table):
        super(MergeTree, self)._set_parent(table)
        if self.partition_by is not None:
            self.partition_by._set_parent(table)
        if self.order_by is not None:
            self.order_by._set_parent(table)
        if self.primary_key is not None:
            self.primary_key._set_parent(table)
        if self.sample_by is not None:
            self.sample_by._set_parent(table)
        if self.ttl is not None:
            self.ttl._set_parent(table)

    @classmethod
    def _reflect_merge_tree(
            self, partition_key=None, sorting_key=None, primary_key=None,
            sampling_key=None, ttl=None, **kwargs):

        # TODO: reflect settings
        rv = {}
        if partition_key:
            rv['partition_by'] = parse_columns(partition_key)
        if sorting_key:
            rv['order_by'] = parse_columns(sorting_key)
        if primary_key:
            rv['primary_key'] = parse_columns(primary_key)
        if sampling_key:
            rv['sample_by'] = parse_columns(sampling_key)
        if ttl:
            rv['ttl'] = parse_columns(ttl)

        return rv

    @classmethod
    def reflect(cls, engine_full, **kwargs):
        return cls(**cls._reflect_merge_tree(**kwargs))


class AggregatingMergeTree(MergeTree):
    pass


class GraphiteMergeTree(MergeTree):

    def __init__(self, config_name, *args, **kwargs):
        super(GraphiteMergeTree, self).__init__(*args, **kwargs)
        self.config_name = config_name

    def get_parameters(self):
        return "'{}'".format(self.config_name)

    @classmethod
    def reflect(cls, engine_full, **kwargs):
        engine = parse_columns(engine_full, delimeter=' ')[0]
        config_name = engine[len(cls.__name__):].strip("()'")

        return cls(
            config_name,
            **cls._reflect_merge_tree(**kwargs)
        )


class CollapsingMergeTree(MergeTree):
    def __init__(self, sign_col, *args, **kwargs):
        super(CollapsingMergeTree, self).__init__(*args, **kwargs)
        self.sign_col = TableCol(sign_col)

    def get_parameters(self):
        return self.sign_col.get_column()

    def _set_parent(self, table):
        super(CollapsingMergeTree, self)._set_parent(table)

        self.sign_col._set_parent(table)

    @classmethod
    def reflect(cls, engine_full, **kwargs):
        engine = parse_columns(engine_full, delimeter=' ')[0]
        sign_col = engine[len(cls.__name__):].strip('()')

        return cls(
            sign_col,
            **cls._reflect_merge_tree(**kwargs)
        )


class VersionedCollapsingMergeTree(MergeTree):
    def __init__(self, sign_col, version_col, *args, **kwargs):
        super(VersionedCollapsingMergeTree, self).__init__(*args, **kwargs)

        self.sign_col = TableCol(sign_col)
        self.version_col = TableCol(version_col)

    def get_parameters(self):
        return [self.sign_col.get_column(), self.version_col.get_column()]

    def _set_parent(self, table):
        super(VersionedCollapsingMergeTree, self)._set_parent(table)

        self.sign_col._set_parent(table)
        self.version_col._set_parent(table)

    @classmethod
    def reflect(cls, engine_full, **kwargs):
        engine = parse_columns(engine_full, delimeter=' ')[0]
        columns = engine[len(cls.__name__):].strip('()')
        sign_col, version_col = parse_columns(columns)

        return cls(
            sign_col, version_col,
            **cls._reflect_merge_tree(**kwargs)
        )


class SummingMergeTree(MergeTree):
    def __init__(self, *args, **kwargs):
        summing_cols = kwargs.pop('columns', None)
        super(SummingMergeTree, self).__init__(*args, **kwargs)

        self.summing_cols = None
        if summing_cols is not None:
            self.summing_cols = KeysExpressionOrColumn(*to_list(summing_cols))

    def _set_parent(self, table):
        super(SummingMergeTree, self)._set_parent(table)

        if self.summing_cols is not None:
            self.summing_cols._set_parent(table)

    def get_parameters(self):
        if self.summing_cols is not None:
            cols = self.summing_cols.get_expressions_or_columns()
            return [cols] if len(cols) > 1 else cols

    @classmethod
    def reflect(cls, engine_full, **kwargs):
        engine = parse_columns(engine_full, delimeter=' ')[0]
        columns = engine[len(cls.__name__):].strip('()')
        columns = parse_columns(columns) or None

        return cls(
            columns=columns,
            **cls._reflect_merge_tree(**kwargs)
        )


class ReplacingMergeTree(MergeTree):
    def __init__(self, *args, **kwargs):
        version_col = kwargs.pop('version', None)
        super(ReplacingMergeTree, self).__init__(*args, **kwargs)

        self.version_col = None
        if version_col is not None:
            self.version_col = TableCol(version_col)

    def _set_parent(self, table):
        super(ReplacingMergeTree, self)._set_parent(table)

        if self.version_col is not None:
            self.version_col._set_parent(table)

    def get_parameters(self):
        if self.version_col is not None:
            return self.version_col.get_column()

    @classmethod
    def reflect(cls, engine_full, **kwargs):
        engine = parse_columns(engine_full, delimeter=' ')[0]
        version_col = engine[len(cls.__name__):].strip('()') or None

        return cls(
            version=version_col,
            **cls._reflect_merge_tree(**kwargs)
        )




from .mergetree import (
    MergeTree, AggregatingMergeTree, GraphiteMergeTree, CollapsingMergeTree,
    VersionedCollapsingMergeTree, ReplacingMergeTree, SummingMergeTree
)
from .misc import Distributed, Buffer, TinyLog, Log, Memory, Null, File
from .replicated import (
    ReplicatedMergeTree, ReplicatedAggregatingMergeTree,
    ReplicatedCollapsingMergeTree, ReplicatedVersionedCollapsingMergeTree,
    ReplicatedReplacingMergeTree, ReplicatedSummingMergeTree
)


__all__ = (
    MergeTree,
    AggregatingMergeTree,
    GraphiteMergeTree,
    CollapsingMergeTree,
    VersionedCollapsingMergeTree,
    SummingMergeTree,
    ReplacingMergeTree,
    Distributed,
    ReplicatedMergeTree,
    ReplicatedAggregatingMergeTree,
    ReplicatedCollapsingMergeTree,
    ReplicatedVersionedCollapsingMergeTree,
    ReplicatedReplacingMergeTree,
    ReplicatedSummingMergeTree,
    Buffer,
    TinyLog,
    Log,
    Memory,
    Null,
    File
)




from .base import Engine
from .util import parse_columns


class Distributed(Engine):
    def __init__(self, logs, default, hits, sharding_key=None):
        self.logs = logs
        self.default = default
        self.hits = hits
        self.sharding_key = sharding_key
        super(Distributed, self).__init__()

    def get_parameters(self):
        params = [
            self.logs,
            self.default,
            self.hits
        ]

        if self.sharding_key is not None:
            params.append(self.sharding_key)

        return params

    @classmethod
    def reflect(cls, engine_full, **kwargs):
        engine = parse_columns(engine_full, delimeter=' ')[0]
        columns = engine.split('(', 1)[1][:-1]
        columns = parse_columns(columns)

        return cls(*columns)


class Buffer(Engine):
    def __init__(self, database, table, num_layers=16,
                 min_time=10, max_time=100, min_rows=10000, max_rows=1000000,
                 min_bytes=10000000, max_bytes=100000000):
        self.database = database
        self.table_name = table
        self.num_layers = num_layers
        self.min_time = min_time
        self.max_time = max_time
        self.min_rows = min_rows
        self.max_rows = max_rows
        self.min_bytes = min_bytes
        self.max_bytes = max_bytes
        super(Buffer, self).__init__()

    def get_parameters(self):
        return [
            self.database,
            self.table_name,
            self.num_layers,
            self.min_time,
            self.max_time,
            self.min_rows,
            self.max_rows,
            self.min_bytes,
            self.max_bytes
        ]

    @classmethod
    def reflect(cls, engine_full, **kwargs):
        engine = parse_columns(engine_full, delimeter=' ')[0]
        params = parse_columns(engine[len(cls.__name__):].strip("()"))

        database = params[0]
        table_name = params[1]
        num_layers = int(params[2])
        min_time = int(params[3])
        max_time = int(params[4])
        min_rows = int(params[5])
        max_rows = int(params[6])
        min_bytes = int(params[7])
        max_bytes = int(params[8])

        return cls(
            database, table_name, num_layers, min_time, max_time,
            min_rows, max_rows, min_bytes, max_bytes
        )


class _NoParamsEngine(Engine):
    @classmethod
    def reflect(cls, engine_full, **kwargs):
        return cls()


class TinyLog(_NoParamsEngine):
    pass


class Log(_NoParamsEngine):
    pass


class Memory(_NoParamsEngine):
    pass


class Null(_NoParamsEngine):
    pass


class File(Engine):
    supported_data_formats = {
        'tabseparated': 'TabSeparated',
        'tabseparatedwithnames': 'TabSeparatedWithNames',
        'tabseparatedwithnamesandtypes': 'TabSeparatedWithNamesAndTypes',
        'template': 'Template',
        'csv': 'CSV',
        'csvwithnames': 'CSVWithNames',
        'customseparated': 'CustomSeparated',
        'values': 'Values',
        'jsoneachrow': 'JSONEachRow',
        'tskv': 'TSKV',
        'protobuf': 'Protobuf',
        'parquet': 'Parquet',
        'rowbinary': 'RowBinary',
        'rowbinarywithnamesandtypes': 'RowBinaryWithNamesAndTypes',
        'native': 'Native'
    }

    def __init__(self, data_format):
        fmt = self.supported_data_formats.get(data_format.lower())
        if not fmt:
            raise ValueError(
                'Format {0} not supported for File engine'.format(
                    data_format
                )
            )
        self.data_format = fmt
        super(File, self).__init__()

    def get_parameters(self):
        return (self.data_format, )

    @classmethod
    def reflect(cls, engine_full, **kwargs):
        fmt = engine_full.split('(', 1)[1][:-1].strip("'")
        return cls(fmt)



from sqlalchemy.sql import ClauseElement
from sqlalchemy.sql.base import SchemaEventTarget
from sqlalchemy.sql.schema import ColumnCollectionMixin, SchemaItem
from sqlalchemy.sql.visitors import Visitable
from sqlalchemy.util import zip_longest


class Engine(SchemaEventTarget, Visitable):
    __visit_name__ = 'engine'

    def get_parameters(self):
        return []

    def extend_parameters(self, *params):
        rv = []
        for param in params:
            if isinstance(param, (tuple, list)):
                rv.extend(param)
            elif param is not None:
                rv.append(param)
        return rv

    @property
    def name(self):
        return self.__class__.__name__

    def _set_parent(self, table):
        self.table = table
        self.table.engine = self

    @classmethod
    def reflect(cls, engine_full, **kwargs):
        raise NotImplementedError


class TableCol(ColumnCollectionMixin, SchemaItem):
    def __init__(self, column, **kwargs):
        super(TableCol, self).__init__(*[column], **kwargs)

    def get_column(self):
        return list(self.columns)[0]


class KeysExpressionOrColumn(ColumnCollectionMixin, SchemaItem):
    def __init__(self, *expressions, **kwargs):
        columns = []
        self.expressions = []
        for expr, column, strname, add_element in self.\
                _extract_col_expression_collection(expressions):
            if add_element is not None:
                columns.append(add_element)
            self.expressions.append(expr)

        super(KeysExpressionOrColumn, self).__init__(*columns, **kwargs)

    def _set_parent(self, table):
        super(KeysExpressionOrColumn, self)._set_parent(table)

    def get_expressions_or_columns(self):
        expr_columns = zip_longest(self.expressions, self.columns)
        return [
            (expr if isinstance(expr, ClauseElement) else colexpr)
            for expr, colexpr in expr_columns
        ]
